#!/usr/bin/env perl

use Data::Dumper;
use DBI;
use Session::Token;
use POSIX qw/strftime locale_h/;
use HTML::Template;
require "/usr/share/libcgi-yazzy/cgi.pl";

setlocale(LC_TIME, 'C');
$ENV{'TZ'} = "GMT";

$tokendb_file = "/var/lib/www-auth/tokens.sqlite";
$token_name = "htpasswd_token";
$domain_cookie = 0;
%TimeMultiply = (s=>1, m=>60, h=>3600, d=>86400, w=>604800);
$datetime_fmt = "%a, %d-%b-%Y %H:%M:%S %Z";
$epoch = strftime($datetime_fmt, localtime 0);
$y2k38 = "Tue, 19-Jan-2038 03:14:08 GMT";
$never_sec = 20*365*24*3600;
$unused_timeo = 10*60;
$clausa_nonexpired = "(CASE used WHEN 0 THEN created + $unused_timeo ELSE used + timeout END) >  strftime('%s', 'now')*1";
$clausa_yesexpired = "(CASE used WHEN 0 THEN created + $unused_timeo ELSE used + timeout END) <= strftime('%s', 'now')*1";


sub error_response
{
	my $str = shift;
	my $txt = shift;
	$\ = $CRLF;
	print "Status: 500";
	print "Content-Type: text/plain";
	print "";
	print "$str\n$txt";
	exit 0;
}

sub datetime_or_never($)
{
	return $_[0] > 0 ? strftime($datetime_fmt, localtime($_[0])) : "never";
}

sub humantime($)
{
	my $t = shift;
	my $unit = "sec";
	for my $a (['min', 60], ['h', 60], ['day', 24], ['year', 365])
	{
		if($t > $a->[1])
		{
			$t /= $a->[1];
			$unit = $a->[0];
		}
		else
		{
			last;
		}
	}
	return (sprintf "%d", $t), $unit;
}

sub db_conn
{
	if(!$db)
	{
		$db = DBI->connect("dbi:SQLite:dbname=$tokendb_file", '', '', {AutoCommit=>1, PrintError=>0, RaiseError=>1});
		if(!$db)
		{
			error_response "Database Access Error";
		}
		if(chmod(0600, $tokendb_file) < 1)
		{
			error_response "Database Mode Error";
		}
	}
	1;
}

sub get_tokens_hash(;$$)
{
	my $str = $_[0] || $ENV{'HTTP_COOKIE'};
	my @tokens;
	for my $cookie (split /\s*;\s*/, $str)
	{
		if($cookie =~ /$token_name(\d*)=(.+)/)
		{
			push @tokens, [$1, $2];
		}
	}
	if(exists $_[1])
	{
		my ($next) = (sort {$b<=>$a} map {$_->[0]} @tokens);
		$next++;
		$_[1] = $next;
	}
	return @tokens;
}
sub get_tokens(;$$)
{
	return map {$_->[1]} &get_tokens_hash(@_);
}



$host = $ENV{'HTTP_HOST'};
$is_secure = $ENV{'HTTPS'} ? "Secure" : "";
$mypath = $ENV{'REQUEST_URI'};


if($ENV{'REQUEST_METHOD'} eq "POST")
{
	if($_POST{'action'} eq "invalidate")
	{
		for my $ref (get_tokens_hash())
		{
			my ($num, $token) = @$ref;
			if($_POST{'token'} eq $token)
			{
				my $user = $_POST{'user'};
				db_conn();
				my ($path) = $db->selectrow_array("SELECT path FROM token WHERE token=? AND user=?", undef, $token, $user);
				$ok = $db->do("DELETE FROM token WHERE token=? AND user=?", undef, $token, $user);
				if($ok)
				{
					my ($more_creds) = $db->selectrow_array("SELECT 1 FROM token WHERE token=?", undef, $token);
					$\ = $CRLF;
					print "Status: 302 Token Removed";
					if(not $more_creds)
					{
						print "Set-Cookie: ${token_name}=; Expires=$epoch; ", $domain_cookie ? "Domain=$host; " : "", $path ? "Path=$path; " : "", "HttpOnly; $is_secure";
						print "Set-Cookie: ${token_name}$num=; Expires=$epoch; ", $domain_cookie ? "Domain=$host; " : "", "Path=$mypath; HttpOnly; $is_secure";
					}
					print "Location: $ENV{'REQUEST_URI'}";
					print "";
					last;
				}
				else
				{
					error_response "Database Operation Error", $db->errstr;
				}
			}
		}
	}
	else
	{
		for my $var (qw/user pass time_sec time time_unit constip append kickoff uri path extra_paths/)
		{
			$$var = $_POST{$var};
		}
		
		if(not length $path)
		{
			error_response "Parameter Error: path";
		}
		
		if($time_sec eq "")
		{
			$m = substr($time_unit,0,1);
			if($TimeMultiply{$m})
			{
				$time *= $TimeMultiply{$m};
			}
			else
			{
				error_response "Parameter Error: time/time_unit";
			}
		}
		else
		{
			$time = $time_sec;
		}
		if($time == 0)
		{
			$time = $never_sec;
		}
		$expires = strftime($datetime_fmt, localtime(time + $time));
		$constip = lc $constip eq 'on';
		$kickoff = lc $kickoff eq 'on';
		$append  = lc $append  eq 'on';
		
		
		db_conn;
		$ok = $db->do("CREATE TABLE IF NOT EXISTS token (
				token TEXT,
				user TEXT,
				pass TEXT,
				path TEXT,
				extra_paths TEXT,
				bind_ip INTEGER,
				ip TEXT,
				agent TEXT,
				created INTEGER,
				used INTEGER DEFAULT 0,
				timeout INTEGER,
				kickoff INTEGER,
				append INTEGER,
				UNIQUE (token, user, path) ON CONFLICT REPLACE)");
		if(!$ok)
		{
			error_response "Database Operation Error", $db->errstr;
		}
		
		my $uniq_token_num;
		my @tokens_num = get_tokens_hash(undef, $uniq_token_num);
		my @tokens = map {$_->[1]} @tokens_num;
		my @expired_tokens = @{ $db->selectall_arrayref("SELECT token, path, extra_paths FROM token WHERE token IN (".join(",", map {"?"} @tokens).") AND $clausa_yesexpired", {Slice=>{}}, @tokens) };
		my ($new_token) = $db->selectrow_array("SELECT token FROM token WHERE path=? AND token IN (".join(",", map {"?"} @tokens).") AND $clausa_nonexpired ORDER BY created DESC", undef, $path, @tokens);
		
		if(defined $new_token)
		{
			my ($num) = sort {$b<=>$a} map {$_->[0]} grep {$_->[1] eq $new_token} @tokens_num;
			if(defined $num)
			{
				$uniq_token_num = $num;
			}
		}
		else
		{
			$new_token = Session::Token->new(entropy => 256)->get;
		}
		$ok = $db->do("INSERT INTO token 
				(token, user, pass, path, extra_paths, bind_ip, ip, agent, created, timeout, kickoff, append) 
				VALUES (?, ?, ?, ?, ?, ?, ?, ?, strftime('%s', 'now')*1, ?, ?, ?)", undef, 
				$new_token, $user, $pass, $path, $extra_paths, $constip, $ENV{'REMOTE_ADDR'}, $ENV{'HTTP_USER_AGENT'}, $time, $kickoff, $append);
		if(!$ok)
		{
			error_response "Database Operation Error", $db->errstr;
		}
		$db->do("DELETE FROM token WHERE token IN (".join(",", map {"?"} @tokens).") AND $clausa_yesexpired", undef, @tokens);
		
		
		$\ = $CRLF;
		print "Status: 302 Token Acquired";
		for my $cookiepath ($path, split ';', $extra_paths)
		{
			print "Set-Cookie: ${token_name}=$new_token; Expires=$y2k38; ", $domain_cookie ? "Domain=$host; " : "", "Path=$cookiepath; HttpOnly; $is_secure";
		}
		print "Set-Cookie: ${token_name}$uniq_token_num=$new_token; Expires=$y2k38; ", $domain_cookie ? "Domain=$host; " : "", "Path=$mypath; HttpOnly; $is_secure";
		for my $ref_cookie (@tokens_num)
		{
			my ($num, $token_cookie) = @$ref_cookie;
			for my $hash_expired (@expired_tokens)
			{
				if($token_cookie eq $hash_expired->{'token'})
				{
					for my $cookiepath ($hash_expired->{'path'}, split ';', $hash_expired->{'extra_paths'})
					{
						print "Set-Cookie: ${token_name}=; Expires=$epoch; ", $domain_cookie ? "Domain=$host; " : "", "Path=$cookiepath; HttpOnly; $is_secure";
					}
					print "Set-Cookie: ${token_name}$num=; Expires=$epoch; ", $domain_cookie ? "Domain=$host; " : "", "Path=$mypath; HttpOnly; $is_secure";
				}
			}
		}
		print "Location: $uri";
		print "";
	}
}
elsif($ENV{'REQUEST_METHOD'} eq "GET")
{
	my @sessions;
	my @tokens = get_tokens();
	if(@tokens)
	{
		db_conn;
		@sessions = @{ $db->selectall_arrayref("SELECT token, user, path, extra_paths, bind_ip, ip, agent, created, used, timeout, ($clausa_yesexpired) AS expired FROM token WHERE token IN (".join(",", map {"?"} @tokens).")", {Slice=>{}}, @tokens) };
	}
	for my $rec (@sessions)
	{
		$rec->{"datetime-created"} = datetime_or_never($rec->{'created'});
		$rec->{"datetime-used"} = datetime_or_never($rec->{'used'});
		$rec->{"expire"} = $rec->{'used'} ? ($rec->{'used'} + $rec->{'timeout'}) : ($rec->{'created'} + $unused_timeo);
		$rec->{"expiresin_sec"} = $rec->{"expire"} - time;
		($rec->{"expiresin"}, $rec->{"expiresin_unit"}) = humantime($rec->{"expire"} - time);
		$rec->{"datetime-expire"} = datetime_or_never($rec->{"expire"});
	}
	for my $token (@tokens)
	{
		unless(grep {$_->{'token'} eq $token} @sessions)
		{
			push @sessions, {token=>$token};
		}
	}

	$\ = $CRLF;
	print "Status: 200 OK";
	print "Content-Type: text/html";
	print "Cache-Control: no-cache";
	print "Expires: Thu, 01-Jan-1970 00:00:00 GMT";
	print "";

	$tmpl = HTML::Template->new(filename => '/usr/share/authd-htpasswd/tokens.htm', die_on_bad_params => 0);
	$tmpl->param(SESSIONS => \@sessions);
	print $tmpl->output;
}

